import json
from typing import Any

from dash.development.base_component import Component
from pydantic import BaseModel
from selenium.webdriver.common.by import By

from dash_pydantic_form import ids
from dash_pydantic_form.utils import get_non_null_annotation


def find_ids(component: Component, ids: list[str | dict], whole_elem: bool = False):
    """Find ids in a component tree."""
    ids_found = []
    ids_not_found = ids
    for elem in component._traverse_ids():
        if elem.id in ids:
            ids_found.append(elem if whole_elem else elem.id)
            ids_not_found.remove(elem.id)
    return ids_found, ids_not_found


def check_ids_exist(component: Component, ids: list[str | dict]):
    """Check if ids exist in a component tree."""
    _, ids_not_found = find_ids(component, ids)
    if ids_not_found:
        raise ValueError(f"Could not find ids: {ids_not_found}")


def check_ids_absent(component: Component, ids: list[str | dict]):
    """Check if ids are absent in a component tree."""
    ids_found, _ = find_ids(component, ids)
    if ids_found:
        raise ValueError(f"Found ids: {ids_found}")


def check_elem_values(component: Component, expected: dict):
    """Check if elements in a component tree have the expected values."""
    mismatched_values = {}
    for elem in component._traverse_ids():
        str_id = json.dumps(elem.id)
        attribute = "value" if ids.value_field.args[0] in str_id else "checked"
        if str_id in expected and getattr(elem, attribute) != expected[str_id]:
            mismatched_values[str_id] = (getattr(elem, attribute), expected[str_id])
    if mismatched_values:
        raise ValueError(f"Mismatched values: {mismatched_values}")


def get_field_ids(model: type[BaseModel], aio_id: str, form_id: str, fields: list[str] | None = None, **id_kwargs):
    """Get field ids for a model."""
    fields = fields or list(model.model_fields)
    for field_name in fields:
        field_info = model.model_fields[field_name]
        if fields is None or field_name in fields:
            base_id = ids.checked_field if get_non_null_annotation(field_info.annotation) is bool else ids.value_field
            yield base_id(aio_id, form_id, field_name, **id_kwargs)


def stringify_id(id: dict):
    """Stringify dict id to use with selenium."""
    if isinstance(id, str):
        return id
    return json.dumps(id, sort_keys=True).replace(" ", "").replace('"', r"\"")


def set_input(dash_duo, id: str | dict, value: Any):
    """Set input value."""
    str_id = stringify_id(id)
    elem = dash_duo.driver.find_element(By.ID, str_id)
    elem.clear()
    elem.send_keys(str(value))


def set_select(dash_duo, id: str | dict, value: str):
    """Set select value."""
    str_id = stringify_id(id)
    dash_duo.driver.find_element(By.ID, str_id).click()
    dash_duo.driver.find_element(
        By.CSS_SELECTOR, f"[aria-labelledby='{str_id}-label'] [role='option'][value='{value}']"
    ).click()


def set_checkbox(dash_duo, id: str | dict, value: bool):
    """Set checkbox value."""
    str_id = stringify_id(id)
    elem = dash_duo.driver.find_element(By.ID, str_id)
    if elem.get_property("checked") != value:
        elem.click()



import logging
from datetime import date, datetime, time
from enum import Enum
from typing import Literal, get_origin

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from dash_pydantic_form.utils import Type, get_non_null_annotation, is_subclass

from . import all_fields as fields
from .base_fields import BaseField, VisibilityFilter

DEFAULT_FIELDS_REPR: dict[type, BaseField] = {
    str: fields.Text,
    int: fields.Number,
    float: fields.Number,
    bool: fields.Checkbox,
    date: fields.Date,
    datetime: fields.Datetime,
    time: fields.Time,
    list: fields.MultiSelect,
    Literal: fields.Select,
    Enum: fields.Select,
    BaseModel: fields.Model,
}
DEFAULT_REPR = fields.Json


def get_default_repr(field_info: FieldInfo | None, annotation: type | None = None, **kwargs) -> BaseField:  # noqa: PLR0911, PLR0912
    """Get default field representation."""
    if field_info is not None:
        # Add default repr kwargs
        if (
            field_info.json_schema_extra
            and (repr_kwargs := field_info.json_schema_extra.get("repr_kwargs")) is not None
        ):
            kwargs = repr_kwargs | kwargs

        # Use repr_type if provided and exists
        if field_info.json_schema_extra and (repr_type := field_info.json_schema_extra.get("repr_type")) is not None:
            repr_cls = getattr(fields, repr_type, None)
            if repr_cls is not None:
                return repr_cls(**kwargs)
            logging.warning("Unknown repr_type: %s", repr_type)

        ann = get_non_null_annotation(field_info.annotation)
        type_ = Type.classify(field_info.annotation, discriminator=field_info.discriminator)
    else:
        if annotation is None:
            raise ValueError("Either field_info or annotation must be provided")
        ann = annotation
        type_ = Type.classify(ann)

    if type_ in [Type.MODEL, Type.DISCRIMINATED_MODEL]:
        return fields.Model(**kwargs)

    if type_ in [Type.MODEL_LIST, Type.SCALAR_LIST]:
        if type_ == Type.SCALAR_LIST:
            kwargs.update(render_type="scalar")
        return fields.List(**kwargs)

    if type_ in [Type.MODEL_DICT, Type.SCALAR_DICT, Type.LITERAL_DICT]:
        if type_ in [Type.SCALAR_DICT, Type.LITERAL_DICT]:
            kwargs.update(render_type="scalar")
        return fields.Dict(**kwargs)

    if type_ == Type.SCALAR and ann in DEFAULT_FIELDS_REPR:
        return DEFAULT_FIELDS_REPR.get(ann, DEFAULT_REPR)(**kwargs)

    # Test for type origin
    origin = get_origin(ann)
    if origin in DEFAULT_FIELDS_REPR:
        return DEFAULT_FIELDS_REPR[origin](**kwargs)

    # Test for subclass
    for type_, field_repr in DEFAULT_FIELDS_REPR.items():
        if is_subclass(ann, type_):
            return field_repr(**kwargs)

    return DEFAULT_REPR(**kwargs)


__all__ = [
    "BaseField",
    "VisibilityFilter",
    "fields",
]



import json
import logging
from collections.abc import Callable
from functools import partial
from typing import ClassVar, Literal

import dash_mantine_components as dmc
from dash import (
    ALL,
    MATCH,
    Input,
    Output,
    State,
    callback,
    clientside_callback,
    ctx,
    dcc,
    no_update,
)
from dash.development.base_component import Component
from dash_iconify import DashIconify
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import BaseField


def side_id(  # noqa: PLR0913
    component: str,
    aio_id: str,
    form_id: str,
    field: str,
    parent: str,
    meta: str,
    side: Literal["left", "right"],
):
    """TransferListField side id."""
    return common_ids.field_dependent_id(component, aio_id, form_id, field, parent, meta) | {"side": side}


class TransferListField(BaseField):
    """TransferListField."""

    base_component = None
    full_width = True

    titles: tuple[str, str] | None = Field(default=None, description="List titles.")
    search_placeholder: str | None = Field(default=None, description="Search placeholder.")
    show_transfer_all: bool = Field(default=True, description="Show transfer all button.")
    list_height: int = Field(default=200, description="List rows height in pixels.")
    max_items: int = Field(default=0, description="Max number of items to display.")
    radius: int | str | None = Field(default=None, description="List border-radius.")
    placeholder: str | None = Field(default=None, description="Placeholder text when no item is available.")
    nothing_found: str | None = Field(default=None, description="Text to display when no item is available.")
    transfer_all_matching_filters: bool = Field(default=True, description="Transfer all when filters match.")
    options_labels: dict | None = Field(default=None, description="Label for the options list.")
    data_getter: str = Field(description="Data getter name, needs to have been registered with register_data_getter.")

    getters: ClassVar[dict[str, Callable[[str | None, int | None], list[str]]]] = {}

    class ids:
        """TransferListField ids."""

        search = partial(side_id, "__trl-search-input")
        transfer = partial(side_id, "__trl-transfer-button")
        transfer_tooltip = partial(side_id, "__trl-transfer-tooltip")
        transfer_all = partial(side_id, "__trl-transfer-all-button")
        transfer_all_tooltip = partial(side_id, "__trl-transfer-all-tooltip")
        checklist = partial(side_id, "__trl-checklist")

    @classmethod
    def register_data_getter(
        cls,
        data_getter: Callable[[str | None, int | None], list[str]],
        name: str | None = None,
    ):
        """Register a data_getter."""
        name = name or str(data_getter)
        if name in cls.getters:
            logging.warning("Data getter %s already registered for TransferList field.", name)
        cls.getters[name] = data_getter

    @classmethod
    def get_data(cls, data_getter: str, value: list, search: str | None = None, max_items: int | None = None) -> dict:
        """Retrieve data from Literal annotation if data is not present in input_kwargs."""
        try:
            getter = cls.getters[data_getter]
        except KeyError as exc:
            logging.error(
                "Data getter %s could not be found, make sure you register it at the root level.", data_getter
            )
            raise exc
        search_data = getter(search, max_items + len(value) if max_items else max_items)
        data = [x for x in search_data if x not in value]
        if max_items:
            data = data[:max_items]

        return data

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,  # noqa: ARG002
    ) -> Component:
        value = self.get_value(item, field, parent) or []
        data_left = self.get_data(self.data_getter, value, search=None, max_items=self.max_items)
        id_data = {"aio_id": aio_id, "form_id": form_id, "field": field, "parent": parent, "meta": ""}
        return dmc.SimpleGrid(
            [
                # First list
                dmc.Stack(
                    [
                        *([dmc.Text(self.titles[0], fw=600)] if self.titles else []),
                        dmc.Paper(
                            [
                                dmc.Group(
                                    [
                                        self.search_input(id_data, "left", placeholder=self.search_placeholder),
                                        self.transfer(id_data, "left", self.show_transfer_all),
                                        *([self.transfer_all(id_data, "left")] if self.show_transfer_all else []),
                                    ],
                                    style={"alignItems": "initial"},
                                    gap=0,
                                ),
                                self.checklist(
                                    id_data, "left", data_left, self.list_height, self.max_items, self.options_labels
                                ),
                            ],
                            radius=self.radius,
                            withBorder=True,
                            style={"overflow": "hidden"},
                        ),
                    ],
                    gap="0.375rem",
                    style={"gridColumn": "span var(--col-2-4)"},
                ),
                # Second list
                dmc.Stack(
                    [
                        *([dmc.Text(self.titles[1], fw=600)] if self.titles else []),
                        dmc.Paper(
                            [
                                dmc.Group(
                                    [
                                        *([self.transfer_all(id_data, "right")] if self.show_transfer_all else []),
                                        self.transfer(id_data, "right", self.show_transfer_all),
                                        self.search_input(id_data, "right", placeholder=self.search_placeholder),
                                    ],
                                    style={"alignItems": "initial"},
                                    gap=0,
                                ),
                                self.checklist(id_data, "right", value, self.list_height, None, self.options_labels),
                            ],
                            radius=self.radius,
                            withBorder=True,
                            style={"overflow": "hidden"},
                        ),
                    ],
                    gap="0.375rem",
                    style={"gridColumn": "span var(--col-2-4)"},
                ),
                # This input holds the actual value as well as some metadata to pass to callbacks
                dmc.JsonInput(
                    id=common_ids.value_field(**(id_data | {"meta": "transferlist"})),
                    style={"display": "none"},
                    value=value,
                    **{
                        "data-placeholder": json.dumps(self.placeholder),
                        "data-nothingfound": json.dumps(self.nothing_found),
                        "data-getter": self.data_getter,
                        "data-maxitems": self.max_items,
                        "data-transferallmatchingfilters": json.dumps(self.transfer_all_matching_filters),
                    },
                ),
            ],
            className="pydantic-form-grid",
            spacing="md",
            **self.input_kwargs,
        )

    @classmethod
    def checkbox(cls, value: dict | str, options_labels: dict | None = None):
        """Checkbox for the checklist.

        :param value: value of the checkbox, dict with keys label and value
        """
        options_labels = options_labels or {}
        if isinstance(value, str):
            value = {"label": options_labels.get(value, value), "value": value}

        return dmc.Checkbox(
            label=value["label"],
            value=value["value"],
            px="0.5rem",
            styles={
                "labelWrapper": {"flex": 1},
                "label": {"cursor": "pointer", "padding": "0.5rem 0"},
                "body": {"alignItems": "center", "gap": "0.5rem"},
            },
            className="transferlist-checkbox",
        )

    @classmethod
    def checklist(  # noqa: PLR0913
        cls,
        id_data: dict,
        side: Literal["left", "right"],
        value: list[dict],
        list_height: int,
        max_items: int | None = None,
        options_labels: dict | None = None,
    ):
        """Checklist in a scrollarea for each list of the transfer.

        :param id_data: dict to put in the id
        :param side: list side
        :param value: list value, list of dicts with keys label and value
        :param list_height: height of the checklist
        :param max_items: the max number of items displayed in the checklist
        """
        if max_items:
            value = value[:max_items]
        return dcc.Loading(
            dmc.ScrollArea(
                dmc.CheckboxGroup(
                    [cls.checkbox(val, options_labels) for val in value] + cls.more_text(max_items, len(value)),
                    py="0.25rem",
                    id=cls.ids.checklist(**id_data, side=side),
                ),
                style={"height": list_height},
                px="0.25rem",
            ),
            custom_spinner=dmc.Loader(),
            delay_show=500,
        )

    @classmethod
    def search_input(cls, id_data: dict, side: Literal["left", "right"], **kwargs):
        """List search input.

        :param id_data: dict to put in the id
        :param side: list side
        :param **kwargs: kwargs to pass to the TextInput
        """
        return dmc.TextInput(
            styles={
                "root": {"flex": 1},
                "input": {
                    "borderTop": "none",
                    "borderLeft": "none",
                    "borderRight": "none",
                },
            },
            radius=0,
            id=cls.ids.search(**id_data, side=side),
            debounce=250,
            **kwargs,
        )

    @classmethod
    def transfer_all(cls, id_data: dict, side: Literal["left", "right"]):
        """Transfer all button.

        :param id_data: dict to put in the id
        :param side: list side
        """
        icon_side = "left" if side == "right" else "right"
        return dmc.Tooltip(
            dmc.Paper(
                dmc.UnstyledButton(
                    DashIconify(icon=f"uiw:d-arrow-{icon_side}", height=12),
                    style={"height": 34, "width": 34, "display": "grid", "placeContent": "center"},
                    id=cls.ids.transfer_all(**id_data, side=side),
                ),
                withBorder=True,
                radius=0,
                h="100%",
                style={
                    "borderTop": "none",
                    "borderLeft": "none",
                    "borderRight": "none",
                },
            ),
            label="Transfer all",
            id=cls.ids.transfer_all_tooltip(**id_data, side=side),
        )

    @classmethod
    def transfer(cls, id_data: dict, side: Literal["left", "right"], show_transfer_all: bool):
        """Transfer button.

        :param id_data: dict to put in the id
        :param side: list side
        :param show_transfer_all: whether the transfer all button is visible (impacts border style)
        """
        icon_side = "left" if side == "right" else "right"
        return dmc.Tooltip(
            dmc.Paper(
                dmc.UnstyledButton(
                    DashIconify(icon=f"uiw:{icon_side}", height=12),
                    style={"height": 34, "width": 34, "display": "grid", "placeContent": "center"},
                    id=cls.ids.transfer(**id_data, side=side),
                ),
                withBorder=True,
                radius=0,
                h="100%",
                style={"borderTop": "none"}
                | ({"borderRight": "none"} if not show_transfer_all and side == "right" else {})
                | ({"borderLeft": "none"} if not show_transfer_all and side == "left" else {}),
            ),
            label="Transfer selected",
            id=cls.ids.transfer_tooltip(**id_data, side=side),
        )

    @classmethod
    def more_text(cls, max_items: int, data_length: int):
        """Text to display when there are more items than max_items."""
        if not max_items or data_length < max_items:
            return []

        return [
            dmc.Text(
                f"Showing first {max_items} items, refine your search for more...",
                size="xs",
                c="dimmed",
                p="0.25rem 0.5rem",
                pos="sticky",
                bottom=0,
                bg="var(--mantine-color-body)",
            ),
        ]


@callback(
    Output(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", MATCH), "children"),
    Output(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", MATCH), "value"),
    Input(TransferListField.ids.search(MATCH, MATCH, MATCH, MATCH, "", MATCH), "value"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "value"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-nothingfound"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-placeholder"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-getter"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-maxitems"),
    State(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", MATCH), "value"),
    prevent_initial_call=True,
)
def filter_checklist(  # noqa: PLR0913
    search: str,
    value: list,
    nothing_found: str,
    placeholder: str,
    data_getter: str,
    max_items: int,
    selection: list[str],
):
    """Filter the list on search."""
    if not ctx.triggered_id:
        return no_update, no_update

    if ctx.triggered_id["side"] == "left":
        filtered = TransferListField.get_data(data_getter, value, search=search, max_items=max_items)
    else:
        filtered = [x for x in value if not search or search.lower() in x.lower()]
    children = None
    updated_selection = [s for s in (selection or []) if s in filtered]
    if filtered:
        children = [TransferListField.checkbox(val) for val in filtered]
        children += TransferListField.more_text(max_items, len(filtered))
    elif search and nothing_found:
        children = dmc.Text(json.loads(nothing_found), p="0.5rem", c="dimmed")
    elif not search and placeholder:
        children = dmc.Text(json.loads(placeholder), p="0.5rem", c="dimmed")
    return children, updated_selection


@callback(
    Output(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "value"),
    Output(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", ALL), "children", allow_duplicate=True),
    Output(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", ALL), "value", allow_duplicate=True),
    Output(TransferListField.ids.search(MATCH, MATCH, MATCH, MATCH, "", ALL), "value", allow_duplicate=True),
    Input(TransferListField.ids.transfer(MATCH, MATCH, MATCH, MATCH, "", ALL), "n_clicks"),
    Input(TransferListField.ids.transfer_all(MATCH, MATCH, MATCH, MATCH, "", ALL), "n_clicks"),
    State(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", ALL), "value"),
    State(TransferListField.ids.search(MATCH, MATCH, MATCH, MATCH, "", ALL), "value"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "value"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-placeholder"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-transferallmatchingfilters"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-getter"),
    State(common_ids.value_field(MATCH, MATCH, MATCH, MATCH, "transferlist"), "data-maxitems"),
    prevent_initial_call=True,
)
def transfer_values(  # noqa: PLR0913
    trigger1: tuple[int, int],
    trigger2: tuple[int, int],
    selection: tuple[list[str], list[str]],
    search: tuple[str, str],
    current_value: list[str],
    placeholder: str,
    transfer_matching: str,
    data_getter: str,
    max_items: int,
):
    """Transfer items from one list to the other."""
    if not (ctx.triggered_id and (any(trigger1) or any(trigger2))):
        return no_update

    side = ctx.triggered_id["side"]
    search_ = search[0] if side == "left" else search[1]
    # Transfer selected items when clicking the transfer button
    if ctx.triggered_id["component"] == TransferListField.ids.transfer("", "", "", "", "", "")["component"]:
        transferred = selection[0] if side == "left" else selection[1]
    # Transfer all items when clicking the transfer all button
    elif search_ and json.loads(transfer_matching):
        # Filter out items that don't match the search if transfer_matching is set
        if side == "left":
            transferred = TransferListField.get_data(data_getter, current_value, search=search_, max_items=0)
        else:
            transferred = [x for x in current_value if search_.lower() in x.lower()]
    else:
        transferred = (
            current_value
            if side == "right"
            else TransferListField.get_data(data_getter, current_value, search=None, max_items=max_items)
        )

    if not transferred:
        return no_update

    # Update the value
    new_value = current_value + transferred if side == "left" else [v for v in current_value if v not in transferred]

    data_left = TransferListField.get_data(data_getter, new_value, search=search[0], max_items=max_items)

    # Create the new checkboxes or placeholder texts
    placeholder = json.loads(placeholder)
    new_children = [
        ([TransferListField.checkbox(v) for v in data_left] + TransferListField.more_text(max_items, len(data_left)))
        if data_left
        else dmc.Text(placeholder, p="0.5rem", c="dimmed"),
        [TransferListField.checkbox(v) for v in new_value if not search[1] or search[1].lower() in v.lower()]
        if new_value
        else dmc.Text(placeholder, p="0.5rem", c="dimmed"),
    ]

    return new_value, new_children, [[]] * 2, [no_update] * 2


# Gray out the transfer button when nothing is selected
clientside_callback(
    """(selection, style) => {
        return [{
            ...style,
            color: !!selection.length ? null : "gray",
            cursor: !!selection.length ? "pointer" : "default",
        }, !selection.length]
    }""",
    Output(TransferListField.ids.transfer(MATCH, MATCH, MATCH, MATCH, "", MATCH), "style"),
    Output(TransferListField.ids.transfer_tooltip(MATCH, MATCH, MATCH, MATCH, "", MATCH), "disabled"),
    Input(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", MATCH), "value"),
    State(TransferListField.ids.transfer(MATCH, MATCH, MATCH, MATCH, "", MATCH), "style"),
)


# Gray out the transfer all button when nothing can be transferred
clientside_callback(
    """(filtered, style) => {
        const disabled = !filtered || !filtered.length
        return [{
            ...style,
            color: !disabled ? null : "gray",
            cursor: !disabled ? "pointer" : "default",
        }, disabled]
    }""",
    Output(TransferListField.ids.transfer_all(MATCH, MATCH, MATCH, MATCH, "", MATCH), "style"),
    Output(TransferListField.ids.transfer_all_tooltip(MATCH, MATCH, MATCH, MATCH, "", MATCH), "disabled"),
    Input(TransferListField.ids.checklist(MATCH, MATCH, MATCH, MATCH, "", MATCH), "children"),
    State(TransferListField.ids.transfer_all(MATCH, MATCH, MATCH, MATCH, "", MATCH), "style"),
)



from functools import partial
from typing import Literal

import dash_mantine_components as dmc
from dash import (
    MATCH,
    ClientsideFunction,
    Input,
    Output,
    clientside_callback,
    html,
)
from dash.development.base_component import Component
from dash_iconify import DashIconify
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo

from dash_pydantic_form import ids as common_ids
from dash_pydantic_form.fields.base_fields import BaseField
from dash_pydantic_form.form_section import Sections
from dash_pydantic_form.utils import get_fullpath


class ModelField(BaseField):
    """Model field, used for nested BaseModel.

    Optional attributes:
    * render_type (one of 'accordion', 'modal', default 'accordion')
    * fields_repr, mapping between field name and field representation
    * sections, list of FormSection for the NestedModel form
    """

    render_type: Literal["accordion", "modal"] = Field(
        default="accordion", description="How to render the model field, one of 'accordion', 'modal'."
    )
    fields_repr: dict[str, dict | BaseField] | None = Field(
        default=None,
        description="Fields representation, mapping between field name and field representation for the nested fields.",
    )
    sections: Sections | None = Field(default=None, description="Sub-form sections.")

    full_width = True

    def model_post_init(self, _context):
        """Model post init."""
        super().model_post_init(_context)
        if self.fields_repr is None:
            self.fields_repr = {}

    class ids(BaseField.ids):
        """Model field ids."""

        form_wrapper = partial(common_ids.field_dependent_id, "_pydf-model-form-wrapper")
        edit = partial(common_ids.field_dependent_id, "_pydf-model-field-edit")
        modal = partial(common_ids.field_dependent_id, "_pydf-model-field-modal")
        modal_save = partial(common_ids.field_dependent_id, "_pydf-model-field-modal-save")

    def modal_item(  # noqa: PLR0913
        self,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Model field modal render."""
        from dash_pydantic_form import ModelForm

        title = self.get_title(field_info, field_name=field)
        new_parent = get_fullpath(parent, field)
        return dmc.Paper(
            dmc.Group(
                [
                    dmc.Text(
                        title,
                        style={
                            "flex": 1,
                            "overflow": "hidden",
                            "textOverflow": "ellipsis",
                            "whiteSpace": "nowrap",
                        },
                    ),
                    dmc.Group(
                        [
                            dmc.ActionIcon(
                                DashIconify(icon="carbon:edit", height=16),
                                variant="light",
                                size="sm",
                                id=self.ids.edit(aio_id, form_id, field, parent=parent),
                            ),
                        ],
                        gap="0.25rem",
                    ),
                    dmc.Modal(
                        [
                            ModelForm(
                                item=item,
                                aio_id=aio_id,
                                form_id=form_id,
                                path=new_parent,
                                fields_repr=self.fields_repr,
                                sections=self.sections,
                                read_only=self.read_only,
                            ),
                            dmc.Group(
                                dmc.Button(
                                    "Save",
                                    leftSection=DashIconify(icon="carbon:save"),
                                    id=self.ids.modal_save(aio_id, form_id, field, parent=parent),
                                ),
                                justify="right",
                                mt="sm",
                            ),
                        ],
                        title=title,
                        id=self.ids.modal(aio_id, form_id, field, parent=parent),
                        style={"--modal-size": "min(calc(100vw - 4rem), 1150px)"},
                    ),
                ],
                gap="sm",
                align="top",
            ),
            withBorder=True,
            radius="sm",
            p="xs",
            className="pydf-model-list-modal-item",
        )

    def accordion_item(  # noqa: PLR0913
        self,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Model field accordion item render."""
        from dash_pydantic_form import ModelForm

        title = self.get_title(field_info, field_name=field)
        description = self.get_description(field_info)
        return dmc.Accordion(
            dmc.AccordionItem(
                value="item",
                children=[
                    dmc.AccordionControl(dmc.Text(title)),
                    dmc.AccordionPanel(
                        [
                            *([dmc.Text(description, size="xs", c="dimmed")] * bool(title) * bool(description)),
                            html.Div(
                                ModelForm(
                                    item=item,
                                    aio_id=aio_id,
                                    form_id=form_id,
                                    path=get_fullpath(parent, field),
                                    fields_repr=self.fields_repr,
                                    sections=self.sections,
                                    discriminator=field_info.discriminator,
                                    read_only=self.read_only,
                                ),
                                id=self.ids.form_wrapper(
                                    aio_id, form_id, field_info.discriminator or "", parent=get_fullpath(parent, field)
                                ),
                            ),
                        ],
                    ),
                ],
            ),
            value="item",
            styles={
                "control": {"padding": "0.5rem"},
                "label": {"padding": 0},
                "item": {
                    "border": "1px solid color-mix(in srgb, var(--mantine-color-gray-light), transparent 40%)",
                    "background": "color-mix(in srgb, var(--mantine-color-gray-light), transparent 80%)",
                    "marginBottom": "0.5rem",
                    "borderRadius": "0.25rem",
                },
                "content": {
                    "display": "flex",
                    "flexDirection": "column",
                    "gap": "0.375rem",
                    "padding": "0.125rem 0.5rem 0.5rem",
                },
            },
            style={"gridColumn": "span var(--col-4-4)"},
        )

    def _render(  # noqa: PLR0913
        self,
        *,
        item: BaseModel,
        aio_id: str,
        form_id: str,
        field: str,
        parent: str = "",
        field_info: FieldInfo | None = None,
    ) -> Component:
        """Create a form field of type checklist to interact with the model field."""
        if self.render_type == "accordion":
            input_ = self.accordion_item(item, aio_id, form_id, field, parent, field_info)
        elif self.render_type == "modal":
            input_ = self.modal_item(item, aio_id, form_id, field, parent, field_info)
        else:
            raise ValueError("Unknown render type.")
        return input_

    # Open a model modal when editing an item
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="syncTrue"),
        Output(ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened", allow_duplicate=True),
        Input(ids.edit(MATCH, MATCH, MATCH, MATCH, MATCH), "n_clicks"),
        prevent_initial_call=True,
    )

    # Close a model modal when saving an item
    clientside_callback(
        ClientsideFunction(namespace="pydf", function_name="syncFalse"),
        Output(ids.modal(MATCH, MATCH, MATCH, MATCH, MATCH), "opened", allow_duplicate=True),
        Input(ids.modal_save(MATCH, MATCH, MATCH, MATCH, MATCH), "n_clicks"),
        prevent_initial_call=True,
    )



from copy import deepcopy

import dash._callback
import pytest
from dash import _dash_renderer

PYDF_CALLBACK_LIST = []
PYDF_CALLBACK_MAP = {}


@pytest.fixture(scope="session", autouse=True)
def init_test_session():
    """Set react version to 18.2.0 to work with DMC."""
    _dash_renderer._set_react_version("18.2.0")
    global PYDF_CALLBACK_LIST, PYDF_CALLBACK_MAP  # noqa: PLW0603
    PYDF_CALLBACK_LIST = deepcopy(dash._callback.GLOBAL_CALLBACK_LIST)
    PYDF_CALLBACK_MAP = deepcopy(dash._callback.GLOBAL_CALLBACK_MAP)


@pytest.fixture(scope="function", autouse=True)
def reset_callbacks():
    """Reload dash pydantic form to add the callbacks."""
    dash._callback.GLOBAL_CALLBACK_LIST = deepcopy(PYDF_CALLBACK_LIST)
    dash._callback.GLOBAL_CALLBACK_MAP = deepcopy(PYDF_CALLBACK_MAP)



from typing import Literal

from dash_iconify import DashIconify
from pydantic import BaseModel

SectionRender = Literal["accordion", "tabs", "steps"]
Position = Literal["top", "bottom", "none"]


class FormSection(BaseModel):
    """Form section model.

    Parameters
    ----------
    name: str
        Section name.
    fields: list[str]
        List of field names from the pydantic model.
    icon: str | DashIconify | None
        Section icon displayed before the section name, optional.
    default_open: bool
        Whether the section is open by default, only used for accordion sections.
    """

    name: str
    fields: list[str]
    icon: str | DashIconify | None = None
    default_open: bool = False

    model_config = {"arbitrary_types_allowed": True}


class Sections(BaseModel):
    """Form sections model.

    Parameters
    ----------
    sections: list[FormSection]
        List of FormSection.
    remaining_fields_position: Literal["top", "bottom", "none"]
        Position of the fields not listed in the sections. Default "top".
    render: Literal["accordion", "tabs", "steps"]
        how the sections should be rendered. Possible values: "accordion", "tabs", "steps".
        Default "accordion".
    render_kwargs: dict | None
        Additional render kwargs passed to the section render functions, optional.
        See `ModelForm.render_accordion_sections`, `ModelForm.render_tabs_sections` and
        `ModelForm.render_steps_sections`
    """

    sections: list[FormSection]
    remaining_fields_position: Position = "top"
    render: SectionRender = "accordion"
    render_kwargs: dict | None = None

    def model_post_init(self, _context):
        """Model post init."""
        if self.render_kwargs is None:
            self.render_kwargs = {}



import json
from datetime import date, time

import dash_mantine_components as dmc
from dash import Dash, Input, Output
from pydantic import BaseModel, Field
from selenium.webdriver.common.by import By

from dash_pydantic_form import ModelForm, fields, ids
from tests.utils import (
    check_ids_exist,
    set_input,
    stringify_id,
)


def test_li0001_basic_list(dash_duo):
    """Test a basic list."""

    class Basic(BaseModel):
        a: list[int] = Field(title="List", default_factory=list)

    app = Dash(__name__)
    aio_id = "aio"
    form_id = "form"
    app.layout = dmc.MantineProvider(
        [
            ModelForm(Basic, aio_id=aio_id, form_id=form_id),
            dmc.Text(id="output"),
        ],
    )
    check_ids_exist(
        app.layout,
        [fields.List.ids.add(aio_id, form_id, "a")],
    )

    @app.callback(
        Output("output", "children"),
        Input(ModelForm.ids.main(aio_id, form_id), "data"),
    )
    def display(form_data):
        return json.dumps(form_data)

    dash_duo.start_server(app)

    dash_duo.wait_for_text_to_equal("#output", "{}")
    add_btn = dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.add(aio_id, form_id, "a")))
    add_btn.click()
    set_input(dash_duo, ids.value_field(aio_id, form_id, "0", parent="a"), 123)
    dash_duo.wait_for_text_to_equal("#output", '{"a": [123]}')
    add_btn.click()
    dash_duo.wait_for_text_to_equal("#output", '{"a": [123, null]}')
    set_input(dash_duo, ids.value_field(aio_id, form_id, "1", parent="a"), 456)
    dash_duo.wait_for_text_to_equal("#output", '{"a": [123, 456]}')

    del_btn0 = dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.delete(aio_id, form_id, "a", meta="0")))
    # 2 clicks required, on to bring to the foreground one to actually click
    del_btn0.click()
    del_btn0.click()
    dash_duo.wait_for_text_to_equal("#output", '{"a": [456]}')


def test_li0002_list_all_scalar_types(dash_duo):
    """Test a list of all scalar types."""

    class Basic(BaseModel):
        a: list[int]
        b: list[str]
        c: list[date]
        d: list[time]

    app = Dash(__name__)
    item = Basic(
        a=[1, 2, 3],
        b=["a", "b", "c"],
        c=[date(2020, 1, 1), date(2020, 1, 2), date(2020, 1, 3)],
        d=[time(12, 0), time(13, 0), time(14, 0)],
    )
    aio_id = "aio"
    form_id = "form"
    app.layout = dmc.MantineProvider(
        [
            ModelForm(item, aio_id=aio_id, form_id=form_id),
            dmc.Text(id="output"),
        ],
    )

    @app.callback(
        Output("output", "children"),
        Input(ModelForm.ids.main(aio_id, form_id), "data"),
    )
    def display(form_data):
        return json.dumps(form_data)

    dash_duo.start_server(app)

    dash_duo.wait_for_text_to_equal(
        "#output",
        '{"a": [1, 2, 3], "b": ["a", "b", "c"], "c": ["2020-01-01", "2020-01-02", "2020-01-03"], '
        '"d": ["2000-01-01T12:00:00", "2000-01-01T13:00:00", "2000-01-01T14:00:00"]}',
    )


def test_li0003_model_list(dash_duo):
    """Test a model list."""

    class Nested2(BaseModel):
        bb: str = ""

    class Nested(BaseModel):
        aa: list[str] = Field(default_factory=list)

    class Basic(BaseModel):
        a: list[Nested] = Field(default_factory=list)
        b: Nested2 = Field(default_factory=Nested2)

    app = Dash(__name__)
    item = Basic()
    aio_id = "aio"
    form_id = "form"
    app.layout = dmc.MantineProvider(
        [
            ModelForm(item, aio_id=aio_id, form_id=form_id),
            dmc.Text(id="output"),
        ],
    )

    @app.callback(
        Output("output", "children"),
        Input(ModelForm.ids.main(aio_id, form_id), "data"),
    )
    def display(form_data):
        return json.dumps(form_data, sort_keys=True)

    dash_duo.start_server(app)

    dash_duo.wait_for_text_to_equal("#output", '{"b": {"bb": ""}}')

    dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.add(aio_id, form_id, "a"))).click()
    dash_duo.driver.find_element(By.CSS_SELECTOR, ".mantine-Accordion-control").click()
    dash_duo.driver.find_element(By.ID, stringify_id(fields.List.ids.add(aio_id, form_id, "aa", parent="a:0"))).click()
    set_input(dash_duo, ids.value_field(aio_id, form_id, "0", parent="a:0:aa"), 1)
    dash_duo.wait_for_text_to_equal("#output", '{"a": [{"aa": ["1"]}], "b": {"bb": ""}}')



import json
from datetime import date
from enum import Enum
from typing import Literal

import dash_mantine_components as dmc
from dash import Dash, Input, Output
from pydantic import BaseModel, Field
from selenium.webdriver.common.by import By

from dash_pydantic_form import ModelForm, fields, ids
from tests.utils import (
    check_elem_values,
    check_ids_exist,
    find_ids,
    get_field_ids,
    set_checkbox,
    set_input,
    set_select,
    stringify_id,
)


class E(Enum):
    """Test enum."""

    A = "A"
    B = "B"


class Basic(BaseModel):
    """Basic model."""

    a: int = Field(title="Field A")
    b: str = Field(title="Field A")
    c: Literal["a", "b"] = Field(title="Field C")
    d: bool = Field(title="Field D")
    e: E = Field(title="Field E")
    f: date = Field(title="Field F")


basic_data = {"a": 1, "b": "foo", "c": "a", "d": True, "e": "B", "f": "2022-01-01"}


def test_bf0001_basic_form():
    """Test a basic form."""
    aio_id = "basic"
    form_id = "form"
    form = ModelForm(Basic, aio_id=aio_id, form_id=form_id)
    check_ids_exist(form, list(get_field_ids(Basic, aio_id, form_id)))
    check_elem_values(
        form,
        {
            json.dumps(fid): None if fid["component"] == ids.value_field.args[0] else False
            for fid in get_field_ids(Basic, aio_id, form_id)
        },
    )

    assert list(basic_data) == list(Basic.model_fields)
    item = Basic(**basic_data)
    form_filled = ModelForm(item, aio_id=aio_id, form_id=form_id)
    check_ids_exist(form_filled, list(get_field_ids(Basic, aio_id, form_id)))
    check_elem_values(
        form_filled,
        {
            json.dumps(fid): val
            for fid, val in zip(get_field_ids(Basic, aio_id, form_id), item.model_dump().values(), strict=True)
        },
    )


def test_bf0002_basic_form_in_browser(dash_duo):
    """Test a basic form in browser."""
    app = Dash(__name__)
    item = Basic(**basic_data)
    aio_id = "basic"
    form_id = "form"
    app.layout = dmc.MantineProvider(ModelForm(item, aio_id=aio_id, form_id=form_id))

    dash_duo.start_server(app)
    for fid, val in zip(get_field_ids(Basic, aio_id, form_id), item.model_dump(mode="json").values(), strict=True):
        str_id = stringify_id(fid)
        elem = dash_duo.driver.find_element(By.ID, str_id)
        if ids.value_field.args[0] in str_id:
            assert elem.get_property("value") == str(val)
        else:
            assert elem.get_property("checked") == val


def test_bf0003_basic_form_form_data(dash_duo):
    """Test a basic form, retrieving its form data."""
    app = Dash(__name__)
    item = Basic(**basic_data)
    aio_id = "basic"
    form_id = "form"
    app.layout = dmc.MantineProvider(
        [
            ModelForm(item, aio_id=aio_id, form_id=form_id),
            dmc.Text(id="output"),
        ]
    )

    @app.callback(
        Output("output", "children"),
        Input(ModelForm.ids.main(aio_id, form_id), "data"),
    )
    def display(form_data):
        return json.dumps(form_data, sort_keys=True)

    dash_duo.start_server(app)

    set_input(dash_duo, ids.value_field(aio_id, form_id, "a"), basic_data["a"])
    dash_duo.wait_for_text_to_equal("#output", json.dumps(basic_data, sort_keys=True))

    set_input(dash_duo, ids.value_field(aio_id, form_id, "a"), 123)
    set_select(dash_duo, ids.value_field(aio_id, form_id, "c"), "b")
    set_checkbox(dash_duo, ids.checked_field(aio_id, form_id, "d"), False)
    dash_duo.wait_for_text_to_equal(
        "#output", json.dumps(basic_data | {"a": 123, "c": "b", "d": False}, sort_keys=True)
    )


def test_bf0005_use_repr_types():
    """Test using repr_type and repr_kwargs in model."""

    class Basic2(BaseModel):
        """Basic model."""

        a: int = Field(title="Field A", repr_kwargs={"placeholder": "Some placeholder"})
        b: Literal["a", "b"] = Field(
            title="Field B",
            json_schema_extra={"repr_type": "RadioItems", "repr_kwargs": {"options_labels": {"a": "A", "b": "B"}}},
        )
        c: bool = Field(title="Field C", json_schema_extra={"repr_type": "Switch"})
        d: Literal["fr", "uk"] = Field(
            title="Field D",
            json_schema_extra={"repr_kwargs": {"options_labels": {"fr": "France", "uk": "United Kingdom"}}},
        )
        e: str = Field(title="Field E", json_schema_extra={"repr_type": "SomeUnknownType"})

    aio_id = "basic"
    form_id = "form"
    form = ModelForm(
        Basic2,
        aio_id=aio_id,
        form_id=form_id,
        fields_repr={
            "c": fields.Checkbox(),
            "d": {"options_labels": {"fr": "France", "uk": "Grande Bretagne"}},
        },
    )

    # Added repr_kwargs
    matches, _ = find_ids(form, [ids.value_field(aio_id, form_id, "a")], whole_elem=True)
    assert len(matches) == 1
    assert matches[0].placeholder == "Some placeholder"

    # Added non-default repr_type and repr_kwargs
    matches, _ = find_ids(form, [ids.value_field(aio_id, form_id, "b")], whole_elem=True)
    assert len(matches) == 1
    assert isinstance(matches[0], dmc.RadioGroup)
    radio1, radio2 = matches[0].children.children
    assert radio1.label == "A" and radio1.value == "a" and radio2.label == "B" and radio2.value == "b"

    # repr_type overridden in ModelForm
    matches, _ = find_ids(form, [ids.checked_field(aio_id, form_id, "c")], whole_elem=True)
    assert len(matches) == 1
    assert isinstance(matches[0], dmc.Checkbox)

    # repr_kwargs overridden in ModelForm
    matches, _ = find_ids(form, [ids.value_field(aio_id, form_id, "d")], whole_elem=True)
    assert len(matches) == 1
    assert matches[0].data == [{"label": "France", "value": "fr"}, {"label": "Grande Bretagne", "value": "uk"}]

    # Unkown repr types get the default
    matches, _ = find_ids(form, [ids.value_field(aio_id, form_id, "e")], whole_elem=True)
    assert len(matches) == 1
    assert isinstance(matches[0], dmc.TextInput)
